
structure STArray = struct

datatype 'a Cell = Invalid | Value of 'a array;

exception AccessedOldVersion;

type 'a array = 'a Cell Unsynchronized.ref;

fun fromList l = Unsynchronized.ref (Value (Array.fromList l));
fun array (size, v) = Unsynchronized.ref (Value (Array.array (size,v)));
fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
fun sub (Unsynchronized.ref Invalid, idx) = raise AccessedOldVersion |
    sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx);
fun update (aref,idx,v) =
  case aref of
    (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
    (Unsynchronized.ref (Value a)) => (
      aref := Invalid;
      Array.update (a,idx,v);
      Unsynchronized.ref (Value a)
    );

fun length (Unsynchronized.ref Invalid) = raise AccessedOldVersion |
    length (Unsynchronized.ref (Value a)) = Array.length a

fun grow (aref, i, x) = case aref of 
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    let val len=Array.length a;
        val na = Array.array (len+i,x) 
    in
      aref := Invalid;
      Array.copy {src=a, dst=na, di=0};
      Unsynchronized.ref (Value na)
    end
    );

fun shrink (aref, sz) = case aref of
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    if sz > Array.length a then 
      raise Size
    else (
      aref:=Invalid;
      Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
    )
  );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;

end;

structure FArray = struct
  datatype 'a Cell = Value of 'a Array.array | Upd of (int*'a*'a Cell Unsynchronized.ref);

  type 'a array = 'a Cell Unsynchronized.ref;

  fun array (size,v) = Unsynchronized.ref (Value (Array.array (size,v)));
  fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
  fun fromList l = Unsynchronized.ref (Value (Array.fromList l));

  fun sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx) |
      sub (Unsynchronized.ref (Upd (i,v,cr)),idx) =
        if i=idx then v
        else sub (cr,idx);

  fun length (Unsynchronized.ref (Value a)) = Array.length a |
      length (Unsynchronized.ref (Upd (i,v,cr))) = length cr;

  fun realize_aux (aref, v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let
          val len = Array.length a;
          val a' = Array.array (len,v);
        in
          Array.copy {src=a, dst=a', di=0};
          Unsynchronized.ref (Value a')
        end
      ) |
      (Unsynchronized.ref (Upd (i,v,cr))) => (
        let val res=realize_aux (cr,v) in
          case res of
            (Unsynchronized.ref (Value a)) => (Array.update (a,i,v); res)
        end
      );

  fun realize aref =
    case aref of
      (Unsynchronized.ref (Value _)) => aref |
      (Unsynchronized.ref (Upd (i,v,cr))) => realize_aux(aref,v);

  fun update (aref,idx,v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let val nref=Unsynchronized.ref (Value a) in
          aref := Upd (idx,Array.sub(a,idx),nref);
          Array.update (a,idx,v);
          nref
        end
      ) |
      (Unsynchronized.ref (Upd _)) =>
        let val ra = realize_aux(aref,v) in
          case ra of
            (Unsynchronized.ref (Value a)) => Array.update (a,idx,v);
          ra
        end
      ;

  fun grow (aref, inc, x) = case aref of 
    (Unsynchronized.ref (Value a)) => (
      let val len=Array.length a;
          val na = Array.array (len+inc,x) 
      in
        Array.copy {src=a, dst=na, di=0};
        Unsynchronized.ref (Value na)
      end
      )
  | (Unsynchronized.ref (Upd _)) => (
    grow (realize aref, inc, x)
  );

  fun shrink (aref, sz) = case aref of
    (Unsynchronized.ref (Value a)) => (
      if sz > Array.length a then 
        raise Size
      else (
        Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
      )
    ) |
    (Unsynchronized.ref (Upd _)) => (
      shrink (realize aref,sz)
    );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;
end;




structure Ta : sig
  val id : 'a -> 'a
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype num = One | Bit0 of num | Bit1 of num
  datatype color = R | B
  datatype ('a, 'b) rbta = Empty |
    Branch of color * ('a, 'b) rbta * 'a * 'b * ('a, 'b) rbta
  type 'a ord
  val less_eq : 'a ord -> 'a -> 'a -> bool
  val less : 'a ord -> 'a -> 'a -> bool
  type 'a preorder
  val ord_preorder : 'a preorder -> 'a ord
  type 'a order
  val preorder_order : 'a order -> 'a preorder
  type 'a linorder
  val order_linorder : 'a linorder -> 'a order
  datatype ('a, 'b) rbt = Rbt of ('a, 'b) rbta
  datatype ('a, 'b) ta_rule = Rule of 'a * 'b * 'a list
  val lhs : ('a, 'b) ta_rule -> 'a
  val hd : 'a list -> 'a
  val tl : 'a list -> 'a list
  val suc : IntInf.int -> IntInf.int
  val rhsl : ('a, 'b) ta_rule -> 'b
  val rhsq : ('a, 'b) ta_rule -> 'a list
  datatype ('a, 'b) ustate_wrapper = USW1 of 'a | USW2 of 'b
  val map : ('a -> 'b) -> 'a list -> 'b list
  val fold : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val rev : 'a list -> 'a list
  val zip : 'a list -> 'b list -> ('a * 'b) list
  datatype 'a tree = Node of 'a * 'a tree list
  val null : 'a list -> bool
  datatype 'a itself = Type
  val empty : 'a linorder -> ('a, 'b) rbt
  val r_prod : ('a, 'b) ta_rule -> ('c, 'd) ta_rule -> (('a * 'c), 'b) ta_rule
  datatype 'a dlist = Dlist of 'a list
  val foldl : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
  val map_of : 'a equal -> ('a * 'b) list -> 'a -> 'b option
  val ord_nat : IntInf.int ord
  val the : 'a option -> 'a
  val impl_of : 'a linorder -> ('a, 'b) rbt -> ('a, 'b) rbta
  val paint : color -> ('a, 'b) rbta -> ('a, 'b) rbta
  val balance : ('a, 'b) rbta -> 'a -> 'b -> ('a, 'b) rbta -> ('a, 'b) rbta
  val rbt_ins :
    'a ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbta -> ('a, 'b) rbta
  val rbt_insert_with_key :
    'a ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbta -> ('a, 'b) rbta
  val rbt_insert : 'a ord -> 'a -> 'b -> ('a, 'b) rbta -> ('a, 'b) rbta
  val insert : 'a linorder -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_lookup : 'a ord -> ('a, 'b) rbta -> 'a -> 'b option
  val lookup : 'a linorder -> ('a, 'b) rbt -> 'a -> 'b option
  val cart :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      ('d -> ('b -> bool) -> ('e -> 'b -> 'b) -> 'b -> 'b) ->
        (unit -> 'b) -> ('c * 'e -> 'b -> 'b) -> 'a -> 'd -> 'b
  val emptya : 'a dlist
  val member : 'a equal -> 'a list -> 'a -> bool
  val inserta : 'a equal -> 'a -> 'a list -> 'a list
  val list_of_dlist : 'a dlist -> 'a list
  val insertb : 'a equal -> 'a -> 'a dlist -> 'a dlist
  val membera : 'a equal -> 'a dlist -> 'a -> bool
  val remdups : 'a equal -> 'a list -> 'a list
  val it_add :
    ('a -> 'b -> 'c -> 'c) ->
      ('d -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c) ->
        'c -> 'd -> 'c
  val equal_nat : IntInf.int equal
  val preorder_nat : IntInf.int preorder
  val order_nat : IntInf.int order
  val list_all : ('a -> bool) -> 'a list -> bool
  val is_none : 'a option -> bool
  val iti_sel :
    ('a ->
      ('b option -> bool) ->
        ('c -> 'b option -> 'b option) -> 'b option -> 'b option) ->
      'a -> ('c -> 'b option) -> 'b option
  val it_size :
    ('a ->
      (IntInf.int -> bool) ->
        ('b -> IntInf.int -> IntInf.int) -> IntInf.int -> IntInf.int) ->
      'a -> IntInf.int
  val remap_rule : ('a -> 'b) -> ('a, 'c) ta_rule -> ('b, 'c) ta_rule
  type 'a hashable
  val hashcode : 'a hashable -> 'a -> IntInf.int
  val bounded_hashcode : 'a hashable -> IntInf.int -> 'a -> IntInf.int
  val def_hashmap_size : 'a hashable -> 'a itself -> IntInf.int
  val list_all_zip : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool
  val ls_memb : 'a equal -> 'a -> 'a dlist -> bool
  datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list
  datatype ('a, 'b) hashmap = RBT_HM of (IntInf.int, ('a, 'b) assoc_list) rbt
  datatype ('a, 'b, 'c) hashedTa_ext =
    HashedTa_ext of
      ('a, unit) hashmap * ('a, 'b) ta_rule dlist *
        ('b, ('a, 'b) ta_rule dlist) hashmap option *
        ('a, ('a, 'b) ta_rule dlist) hashmap option *
        (('a * 'b), ('a, 'b) ta_rule dlist) hashmap option * 'c
  val hta_idx_f :
    ('a, 'b, 'c) hashedTa_ext -> ('b, ('a, 'b) ta_rule dlist) hashmap option
  val linorder_nat : IntInf.int linorder
  val rm_lookup : 'a linorder -> 'a -> ('a, 'b) rbt -> 'b option
  val impl_ofa : ('a, 'b) assoc_list -> ('a * 'b) list
  val lookupa : 'a equal -> ('a, 'b) assoc_list -> 'a -> 'b option
  val lm_lookup : 'a equal -> 'a -> ('a, 'b) assoc_list -> 'b option
  val lookupb :
    'a equal * 'a hashable ->
      'a -> (IntInf.int, ('a, 'b) assoc_list) rbt -> 'b option
  val impl_of_RBT_HM :
    'a hashable -> ('a, 'b) hashmap -> (IntInf.int, ('a, 'b) assoc_list) rbt
  val hm_lookup : 'a equal * 'a hashable -> 'a -> ('a, 'b) hashmap -> 'b option
  val idx_lookup : ('a -> 'b -> 'c option) -> (unit -> 'c) -> 'a -> 'b -> 'c
  val ls_empty : unit -> 'a dlist
  val hl_idx_lookup :
    'a equal * 'a hashable -> 'a -> ('a, 'b dlist) hashmap -> 'b dlist
  val hta_lookup_f :
    'a equal * 'a hashable ->
      'a -> ('b, 'a, 'c) hashedTa_ext -> ('b, 'a) ta_rule dlist
  val ls_ins : 'a equal -> 'a -> 'a dlist -> 'a dlist
  val it_image_filter :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      (unit -> 'b) -> ('d -> 'b -> 'b) -> ('c -> 'd option) -> 'a -> 'b
  val foldli : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val dlist_iteratei : 'a dlist -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val ls_iteratei : 'a dlist -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val ls_image_filter : 'b equal -> ('a -> 'b option) -> 'a dlist -> 'b dlist
  val faccs :
    'a equal -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> 'b tree -> 'a dlist
  val emptyb : ('a, 'b) assoc_list
  val lm_empty : unit -> ('a, 'b) assoc_list
  val rm_update : 'a linorder -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val fst : 'a * 'b -> 'a
  val snd : 'a * 'b -> 'b
  val update_with_aux :
    'b equal -> 'a -> 'b -> ('a -> 'a) -> ('b * 'a) list -> ('b * 'a) list
  val update_with :
    'a equal ->
      'b -> 'a -> ('b -> 'b) -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val update :
    'a equal -> 'a -> 'b -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val lm_update :
    'a equal -> 'a -> 'b -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val updatea :
    'a equal * 'a hashable ->
      'a -> 'b -> (IntInf.int, ('a, 'b) assoc_list) rbt ->
                    (IntInf.int, ('a, 'b) assoc_list) rbt
  val hm_update :
    'a equal * 'a hashable -> 'a -> 'b -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val rm_iterateoi :
    ('a, 'b) rbta -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val rm_iterateoia :
    'a linorder ->
      ('a, 'b) rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val rm_iteratei :
    'a linorder ->
      ('a, 'b) rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val iteratei :
    ('a, 'b) assoc_list -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val lm_iteratei :
    ('a, 'b) assoc_list -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val iterateia :
    'a linorder ->
      ('a, ('b, 'c) assoc_list) rbt ->
        ('d -> bool) -> ('b * 'c -> 'd -> 'd) -> 'd -> 'd
  val hm_iteratei :
    'a hashable ->
      ('a, 'b) hashmap -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val hm_add :
    'a equal * 'a hashable ->
      ('a, 'b) hashmap -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val rm_sel :
    'a linorder -> ('a, 'b) rbt -> ('a * 'b -> 'c option) -> 'c option
  val lm_sel : ('a, 'b) assoc_list -> ('a * 'b -> 'c option) -> 'c option
  val sel :
    'a hashable ->
      (IntInf.int, ('a, 'b) assoc_list) rbt ->
        ('a * 'b -> 'c option) -> 'c option
  val hm_sel :
    'a hashable -> ('a, 'b) hashmap -> ('a * 'b -> 'c option) -> 'c option
  val s_ins : ('a -> unit -> 'b -> 'c) -> 'a -> 'b -> 'c
  val hs_ins :
    'a equal * 'a hashable -> 'a -> ('a, unit) hashmap -> ('a, unit) hashmap
  val s_sel :
    ('a -> ('b * unit -> 'c option) -> 'c option) ->
      'a -> ('b -> 'c option) -> 'c option
  val hs_sel :
    'a hashable -> ('a, unit) hashmap -> ('a -> 'b option) -> 'b option
  val equal_lista : 'a equal -> 'a list -> 'a list -> bool
  val equal_list : 'a equal -> ('a list) equal
  val size_list : 'a list -> IntInf.int
  val it_inter :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      ('c -> 'd -> bool) -> (unit -> 'b) -> ('c -> 'b -> 'b) -> 'a -> 'd -> 'b
  val it_union :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      ('c -> 'b -> 'b) -> 'a -> 'b -> 'b
  val iti_ball :
    ('a -> (bool -> bool) -> ('b -> bool -> bool) -> bool -> bool) ->
      'a -> ('b -> bool) -> bool
  val rm_empty : 'a linorder -> unit -> ('a, 'b) rbt
  val emptyc : 'a hashable -> unit -> (IntInf.int, ('a, 'b) assoc_list) rbt
  val hm_empty_const : 'a hashable -> ('a, 'b) hashmap
  val hm_empty : 'a hashable -> unit -> ('a, 'b) hashmap
  val hs_empty : 'a hashable -> unit -> ('a, unit) hashmap
  val lh_image_filter :
    'b equal * 'b hashable ->
      ('a -> 'b option) -> 'a dlist -> ('b, unit) hashmap
  val brc_iq :
    'a equal * 'a hashable -> ('a, 'b) ta_rule dlist -> ('a, unit) hashmap
  val hta_Qi : ('a, 'b, 'c) hashedTa_ext -> ('a, unit) hashmap
  val mod_nat : IntInf.int -> IntInf.int -> IntInf.int
  val fifo_empty : unit -> 'a list * 'a list
  val s_memb : ('a -> 'b -> 'c option) -> 'a -> 'b -> bool
  val hs_memb : 'a equal * 'a hashable -> 'a -> ('a, unit) hashmap -> bool
  val s_iteratei :
    ('a -> ('b -> bool) -> ('c * unit -> 'b -> 'b) -> 'b -> 'b) ->
      'a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b
  val hs_iteratei :
    'a hashable ->
      ('a, unit) hashmap -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val hs_size : 'a hashable -> ('a, unit) hashmap -> IntInf.int
  val ls_ball : 'a dlist -> ('a -> bool) -> bool
  val ls_size : 'a dlist -> IntInf.int
  val rs_iteratei :
    'a linorder ->
      ('a, unit) rbt -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val rs_size : 'a linorder -> ('a, unit) rbt -> IntInf.int
  val equal_ta_rulea :
    'a equal -> 'b equal -> ('a, 'b) ta_rule -> ('a, 'b) ta_rule -> bool
  val equal_ta_rule : 'a equal -> 'b equal -> ('a, 'b) ta_rule equal
  val ball_disjoint :
    ('a -> ('b -> bool) -> 'c) -> ('b -> 'd -> bool) -> 'a -> 'd -> 'c
  val lh_disjoint :
    'a equal * 'a hashable -> 'a dlist -> ('a, unit) hashmap -> bool
  val hta_mema :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'a tree -> ('b, 'a, 'c) hashedTa_ext -> bool
  val hta_delta : ('a, 'b, 'c) hashedTa_ext -> ('a, 'b) ta_rule dlist
  val hta_idx_f_update :
    (('a, ('b, 'a) ta_rule dlist) hashmap option ->
      ('a, ('b, 'a) ta_rule dlist) hashmap option) ->
      ('b, 'a, 'c) hashedTa_ext -> ('b, 'a, 'c) hashedTa_ext
  val idx_build_stepfun :
    ('a -> 'b -> 'c option) ->
      ('a -> 'c -> 'b -> 'b) ->
        (unit -> 'c) -> ('d -> 'c -> 'c) -> ('d -> 'a) -> 'd -> 'b -> 'b
  val idx_build :
    (unit -> 'a) ->
      ('b -> 'a -> 'c option) ->
        ('b -> 'c -> 'a -> 'a) ->
          (unit -> 'c) ->
            ('d -> 'c -> 'c) ->
              ('e -> ('a -> bool) -> ('d -> 'a -> 'a) -> 'a -> 'a) ->
                ('d -> 'b) -> 'e -> 'a
  val hll_idx_build :
    'a equal -> 'b equal * 'b hashable ->
      ('a -> 'b) -> 'a dlist -> ('b, 'a dlist) hashmap
  val build_rule_index :
    'a equal -> 'b equal -> 'c equal * 'c hashable ->
      (('a, 'b) ta_rule -> 'c) ->
        ('a, 'b) ta_rule dlist -> ('c, ('a, 'b) ta_rule dlist) hashmap
  val build_rule_index_f :
    'a equal -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist -> ('b, ('a, 'b) ta_rule dlist) hashmap
  val hta_ensure_idx_f :
    'a equal -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val hta_mem :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'a tree -> ('b, 'a, 'c) hashedTa_ext -> bool
  val pa_cond :
    'a hashable -> 'b hashable -> 'c hashable ->
      (('a * 'b), unit) hashmap *
        (('a * 'b) list * (('a * 'b), 'c) ta_rule dlist) ->
        bool
  val hashcode_prod : 'a hashable -> 'b hashable -> 'a * 'b -> IntInf.int
  val bounded_hashcode_prod :
    'a hashable -> 'b hashable -> IntInf.int -> 'a * 'b -> IntInf.int
  val def_hashmap_size_prod :
    'a hashable -> 'b hashable -> ('a * 'b) itself -> IntInf.int
  val hashable_prod : 'a hashable -> 'b hashable -> ('a * 'b) hashable
  val equal_proda : 'a equal -> 'b equal -> 'a * 'b -> 'a * 'b -> bool
  val equal_prod : 'a equal -> 'b equal -> ('a * 'b) equal
  val pa_upd_rule :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      (('a * 'b), unit) hashmap ->
        ('a * 'b) list ->
          ('a * 'b) list -> (('a * 'b), unit) hashmap * ('a * 'b) list
  val hta_idx_s :
    ('a, 'b, 'c) hashedTa_ext -> ('a, ('a, 'b) ta_rule dlist) hashmap option
  val hta_lookup_s :
    'a equal * 'a hashable ->
      'a -> ('a, 'b, 'c) hashedTa_ext -> ('a, 'b) ta_rule dlist
  val ls_ins_dj : 'a equal -> 'a -> 'a dlist -> 'a dlist
  val hta_idx_sf :
    ('a, 'b, 'c) hashedTa_ext ->
      (('a * 'b), ('a, 'b) ta_rule dlist) hashmap option
  val hta_lookup_sf :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'a -> 'b -> ('a, 'b, 'c) hashedTa_ext -> ('a, 'b) ta_rule dlist
  val pa_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext ->
          (('a * 'c), unit) hashmap *
            (('a * 'c) list * (('a * 'c), 'b) ta_rule dlist) ->
            (('a * 'c), unit) hashmap *
              (('a * 'c) list * (('a * 'c), 'b) ta_rule dlist)
  val map_to_nat :
    ('a ->
      (IntInf.int * 'b -> bool) ->
        ('c -> IntInf.int * 'b -> IntInf.int * 'b) ->
          IntInf.int * 'b -> IntInf.int * 'b) ->
      (unit -> 'b) -> ('c -> IntInf.int -> 'b -> 'b) -> 'a -> 'b
  val equal_dlista : 'a equal -> 'a dlist -> 'a dlist -> bool
  val equal_dlist : 'a equal -> 'a dlist equal
  val iflt_image : (('a -> 'b option) -> 'c -> 'd) -> ('a -> 'b) -> 'c -> 'd
  val hs_image_filter :
    'a hashable -> 'b equal * 'b hashable ->
      ('a -> 'b option) -> ('a, unit) hashmap -> ('b, unit) hashmap
  val hs_image :
    'a hashable -> 'b equal * 'b hashable ->
      ('a -> 'b) -> ('a, unit) hashmap -> ('b, unit) hashmap
  val s_ins_dj : ('a -> unit -> 'b -> 'c) -> 'a -> 'b -> 'c
  val hm_update_dj :
    'a equal * 'a hashable -> 'a -> 'b -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val hs_ins_dj :
    'a equal * 'a hashable -> 'a -> ('a, unit) hashmap -> ('a, unit) hashmap
  val hs_inter :
    'a equal * 'a hashable ->
      ('a, unit) hashmap -> ('a, unit) hashmap -> ('a, unit) hashmap
  val hs_union :
    'a equal * 'a hashable ->
      ('a, unit) hashmap -> ('a, unit) hashmap -> ('a, unit) hashmap
  val hhh_cart :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, unit) hashmap -> ('b, unit) hashmap -> (('a * 'b), unit) hashmap
  val brc_cond :
    'a hashable -> 'b hashable ->
      ('a, unit) hashmap * ('a list * (('a, 'b) ta_rule, IntInf.int) hashmap) ->
        bool
  val rqrm_lookup :
    'a equal * 'a hashable -> ('a, 'b dlist) hashmap -> 'a -> 'b dlist
  val hashcode_list : 'a hashable -> 'a list -> IntInf.int
  val hashcode_of_ta_rule :
    'a hashable -> 'b hashable -> ('a, 'b) ta_rule -> IntInf.int
  val hashcode_ta_rule :
    'a hashable -> 'b hashable -> ('a, 'b) ta_rule -> IntInf.int
  val bounded_hashcode_ta_rule :
    'a hashable -> 'b hashable -> IntInf.int -> ('a, 'b) ta_rule -> IntInf.int
  val def_hashmap_size_ta_rule :
    'a hashable -> 'b hashable -> ('a, 'b) ta_rule itself -> IntInf.int
  val hashable_ta_rule : 'a hashable -> 'b hashable -> ('a, 'b) ta_rule hashable
  val brc_inner_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule ->
        ('a, unit) hashmap *
          ('a list * (('a, 'b) ta_rule, IntInf.int) hashmap) ->
          ('a, unit) hashmap *
            ('a list * (('a, 'b) ta_rule, IntInf.int) hashmap)
  val brc_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, ('a, 'b) ta_rule dlist) hashmap ->
        ('a, unit) hashmap *
          ('a list * (('a, 'b) ta_rule, IntInf.int) hashmap) ->
          ('a, unit) hashmap *
            ('a list * (('a, 'b) ta_rule, IntInf.int) hashmap)
  val brec_iqm :
    'a equal * 'a hashable -> 'b hashable ->
      ('a, 'b) ta_rule dlist -> ('a, 'b tree) hashmap
  val init_hta :
    ('a, unit) hashmap -> ('a, 'b) ta_rule dlist -> ('a, 'b, unit) hashedTa_ext
  val it_set_to_list :
    ('a ->
      ('b list -> bool) -> ('b -> 'b list -> 'b list) -> 'b list -> 'b list) ->
      'a -> 'b list
  val hs_to_list : 'a hashable -> ('a, unit) hashmap -> 'a list
  val pa_initial :
    'a equal * 'a hashable -> 'b hashable -> 'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext ->
          (('a * 'c), unit) hashmap *
            (('a * 'c) list * (('a * 'c), 'b) ta_rule dlist)
  val while_option : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a option
  val whilea : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
  val hta_proda :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext -> (('a * 'c), 'b, unit) hashedTa_ext
  val hta_idx_s_update :
    (('a, ('a, 'b) ta_rule dlist) hashmap option ->
      ('a, ('a, 'b) ta_rule dlist) hashmap option) ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val build_rule_index_s :
    'a equal * 'a hashable -> 'b equal ->
      ('a, 'b) ta_rule dlist -> ('a, ('a, 'b) ta_rule dlist) hashmap
  val hta_ensure_idx_s :
    'a equal * 'a hashable -> 'b equal ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val hta_idx_sf_update :
    ((('a * 'b), ('a, 'b) ta_rule dlist) hashmap option ->
      (('a * 'b), ('a, 'b) ta_rule dlist) hashmap option) ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val build_rule_index_sf :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist -> (('a * 'b), ('a, 'b) ta_rule dlist) hashmap
  val hta_ensure_idx_sf :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val hta_prod :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext -> (('a * 'c), 'b, unit) hashedTa_ext
  val hashcode_nat : IntInf.int -> IntInf.int
  val bounded_hashcode_nat : IntInf.int -> IntInf.int -> IntInf.int
  val def_hashmap_size_nat : IntInf.int itself -> IntInf.int
  val hashable_nat : IntInf.int hashable
  val htai_mem :
    IntInf.int tree -> (IntInf.int, IntInf.int, unit) hashedTa_ext -> bool
  val rqrm_add :
    'a equal * 'a hashable -> 'b equal ->
      'a -> 'b -> ('a, 'b dlist) hashmap -> ('a, 'b dlist) hashmap
  val fifo_dequeue : 'a list * 'a list -> 'a * ('a list * 'a list)
  val fifo_enqueue : 'a -> 'a list * 'a list -> 'a list * 'a list
  val fifo_isEmpty : 'a equal -> 'a list * 'a list -> bool
  val lhh_union :
    'a equal * 'a hashable ->
      'a dlist -> ('a, unit) hashmap -> ('a, unit) hashmap
  val brec_cond :
    'a equal ->
      ('a, 'b tree) hashmap *
        (('a list * 'a list) *
          ((('a, 'b) ta_rule, IntInf.int) hashmap * 'a option)) ->
        bool
  val brec_construct_witness :
    'a equal * 'a hashable -> 'b hashable ->
      ('a, 'b tree) hashmap -> ('a, 'b) ta_rule -> 'b tree
  val brec_inner_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, unit) hashmap ->
        ('a, 'b) ta_rule ->
          ('a, 'b tree) hashmap *
            (('a list * 'a list) *
              ((('a, 'b) ta_rule, IntInf.int) hashmap * 'a option)) ->
            ('a, 'b tree) hashmap *
              (('a list * 'a list) *
                ((('a, 'b) ta_rule, IntInf.int) hashmap * 'a option))
  val brec_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, ('a, 'b) ta_rule dlist) hashmap ->
        ('a, unit) hashmap ->
          ('a, 'b tree) hashmap *
            (('a list * 'a list) *
              ((('a, 'b) ta_rule, IntInf.int) hashmap * 'a option)) ->
            ('a, 'b tree) hashmap *
              (('a list * 'a list) *
                ((('a, 'b) ta_rule, IntInf.int) hashmap * 'a option))
  val hta_empty :
    'a hashable -> 'b hashable -> unit -> ('a, 'b, unit) hashedTa_ext
  val ls_image : 'b equal -> ('a -> 'b) -> 'a dlist -> 'b dlist
  val hta_remap :
    'a hashable -> 'b equal * 'b hashable -> 'c equal * 'c hashable ->
      ('a -> 'b) -> ('a, 'c, unit) hashedTa_ext -> ('b, 'c, unit) hashedTa_ext
  val equal_ustate_wrappera :
    'a equal -> 'b equal ->
      ('a, 'b) ustate_wrapper -> ('a, 'b) ustate_wrapper -> bool
  val equal_ustate_wrapper :
    'a equal -> 'b equal -> ('a, 'b) ustate_wrapper equal
  val ls_union : 'a equal -> 'a dlist -> 'a dlist -> 'a dlist
  val ls_union_dj : 'a equal -> 'a dlist -> 'a dlist -> 'a dlist
  val hashcode_ustate_wrapper :
    'a hashable -> 'b hashable -> ('a, 'b) ustate_wrapper -> IntInf.int
  val bounded_hashcode_ustate_wrapper :
    'a hashable -> 'b hashable ->
      IntInf.int -> ('a, 'b) ustate_wrapper -> IntInf.int
  val def_hashmap_size_ustate_wrapper :
    'a hashable -> 'b hashable -> ('a, 'b) ustate_wrapper itself -> IntInf.int
  val hashable_ustate_wrapper :
    'a hashable -> 'b hashable -> ('a, 'b) ustate_wrapper hashable
  val hta_union :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext ->
          (('a, 'c) ustate_wrapper, 'b, unit) hashedTa_ext
  val list_to_ls : 'a equal -> 'a list -> 'a dlist
  val rule_states_l : 'a equal -> ('a, 'b) ta_rule -> 'a dlist
  val it_Union_image :
    ('a -> ('b -> bool) -> ('c -> 'd -> 'd) -> 'd -> 'd) ->
      (unit -> 'd) -> ('e -> 'd -> 'd) -> ('c -> 'e) -> 'a -> 'd
  val llh_Union_image :
    'b equal * 'b hashable -> ('a -> 'b dlist) -> 'a dlist -> ('b, unit) hashmap
  val hta_delta_states :
    'a equal * 'a hashable -> ('a, 'b, 'c) hashedTa_ext -> ('a, unit) hashmap
  val hta_states :
    'a equal * 'a hashable -> ('a, 'b, 'c) hashedTa_ext -> ('a, unit) hashmap
  val hh_map_to_nat :
    'a equal * 'a hashable -> ('a, unit) hashmap -> ('a, IntInf.int) hashmap
  val hta_reindex :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b, unit) hashedTa_ext -> (IntInf.int, 'b, unit) hashedTa_ext
  val htai_prod :
    (IntInf.int, IntInf.int, unit) hashedTa_ext ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext ->
        (IntInf.int, IntInf.int, unit) hashedTa_ext
  val it_set_to_List_enq :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      (unit -> 'b) -> ('c -> 'b -> 'b) -> 'a -> 'b
  val it_set_to_fifo :
    ('a ->
      ('b list * 'b list -> bool) ->
        ('b -> 'b list * 'b list -> 'b list * 'b list) ->
          'b list * 'b list -> 'b list * 'b list) ->
      'a -> 'b list * 'b list
  val hs_to_fifo : 'a hashable -> ('a, unit) hashmap -> 'a list * 'a list
  val build_rqrm :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist -> ('a, ('a, 'b) ta_rule dlist) hashmap
  val hta_add_qi :
    'a equal * 'a hashable -> 'b hashable ->
      'a -> ('a, 'b, unit) hashedTa_ext -> ('a, 'b, unit) hashedTa_ext
  val image_filter_cartesian_product :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      ('d -> ('b -> bool) -> ('e -> 'b -> 'b) -> 'b -> 'b) ->
        (unit -> 'b) ->
          ('f -> 'b -> 'b) -> ('c * 'e -> 'f option) -> 'a -> 'd -> 'b
  val image_filter_cp :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      ('d -> ('b -> bool) -> ('e -> 'b -> 'b) -> 'b -> 'b) ->
        (unit -> 'b) ->
          ('f -> 'b -> 'b) ->
            ('c * 'e -> 'f) -> ('c * 'e -> bool) -> 'a -> 'd -> 'b
  val lll_inj_image_filter_cp :
    'c equal ->
      ('a * 'b -> 'c) -> ('a * 'b -> bool) -> 'a dlist -> 'b dlist -> 'c dlist
  val delta_prod_h :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b) ta_rule dlist ->
        ('c, 'b) ta_rule dlist -> (('a * 'c), 'b) ta_rule dlist
  val hta_prodWR :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'd equal * 'd hashable ->
      ('a, 'b, 'c) hashedTa_ext ->
        ('d, 'b, 'e) hashedTa_ext -> (('a * 'd), 'b, unit) hashedTa_ext
  val hta_reduce :
    'a equal * 'a hashable -> 'b equal ->
      ('a, 'b, 'c) hashedTa_ext ->
        ('a, unit) hashmap -> ('a, 'b, unit) hashedTa_ext
  val htai_empty : unit -> (IntInf.int, IntInf.int, unit) hashedTa_ext
  val htai_union :
    (IntInf.int, IntInf.int, unit) hashedTa_ext ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext ->
        (IntInf.int, IntInf.int, unit) hashedTa_ext
  val brc_rcm_init :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist -> (('a, 'b) ta_rule, IntInf.int) hashmap
  val brc_initial :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist ->
        ('a, unit) hashmap * ('a list * (('a, 'b) ta_rule, IntInf.int) hashmap)
  val htai_add_qi :
    IntInf.int ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext ->
        (IntInf.int, IntInf.int, unit) hashedTa_ext
  val htai_prodWR :
    (IntInf.int, IntInf.int, unit) hashedTa_ext ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext ->
        (IntInf.int, IntInf.int, unit) hashedTa_ext
  val sel_disjoint_witness :
    ('a -> ('b -> 'b option) -> 'c) -> ('b -> 'd -> bool) -> 'a -> 'd -> 'c
  val hh_disjoint_witness :
    'a equal * 'a hashable ->
      ('a, unit) hashmap -> ('a, unit) hashmap -> 'a option
  val brec_initial :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, unit) hashmap ->
        ('a, 'b) ta_rule dlist ->
          ('a, 'b tree) hashmap *
            (('a list * 'a list) *
              ((('a, 'b) ta_rule, IntInf.int) hashmap * 'a option))
  val hta_add_rule :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule ->
        ('a, 'b, unit) hashedTa_ext -> ('a, 'b, unit) hashedTa_ext
  val htai_add_rule :
    (IntInf.int, IntInf.int) ta_rule ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext ->
        (IntInf.int, IntInf.int, unit) hashedTa_ext
  val hta_bwd_reduce :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, unit) hashedTa_ext
  val htai_bwd_reduce :
    (IntInf.int, IntInf.int, unit) hashedTa_ext ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext
  val htai_ensure_idx_f :
    (IntInf.int, IntInf.int, unit) hashedTa_ext ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext
  val htai_ensure_idx_s :
    (IntInf.int, IntInf.int, unit) hashedTa_ext ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext
  val htai_ensure_idx_sf :
    (IntInf.int, IntInf.int, unit) hashedTa_ext ->
      (IntInf.int, IntInf.int, unit) hashedTa_ext
  val hta_is_empty_witness :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> 'b tree option
  val htai_is_empty_witness :
    (IntInf.int, IntInf.int, unit) hashedTa_ext -> IntInf.int tree option
end = struct

fun id x = (fn xa => xa) x;

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype num = One | Bit0 of num | Bit1 of num;

datatype color = R | B;

datatype ('a, 'b) rbta = Empty |
  Branch of color * ('a, 'b) rbta * 'a * 'b * ('a, 'b) rbta;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

type 'a linorder = {order_linorder : 'a order};
val order_linorder = #order_linorder : 'a linorder -> 'a order;

datatype ('a, 'b) rbt = Rbt of ('a, 'b) rbta;

datatype ('a, 'b) ta_rule = Rule of 'a * 'b * 'a list;

fun lhs (Rule (q, l, qs)) = q;

fun hd (x :: xs) = x;

fun tl [] = []
  | tl (x :: xs) = xs;

fun suc n = IntInf.+ (n, (1 : IntInf.int));

fun rhsl (Rule (q, l, qs)) = l;

fun rhsq (Rule (q, l, qs)) = qs;

datatype ('a, 'b) ustate_wrapper = USW1 of 'a | USW2 of 'b;

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs;

fun fold f [] = id
  | fold f (x :: xs) = fold f xs o f x;

fun rev xs = fold (fn a => fn b => a :: b) xs [];

fun zip (x :: xs) (y :: ys) = (x, y) :: zip xs ys
  | zip xs [] = []
  | zip [] ys = [];

datatype 'a tree = Node of 'a * 'a tree list;

fun null [] = true
  | null (x :: xs) = false;

datatype 'a itself = Type;

fun empty A_ = Rbt Empty;

fun r_prod (Rule (q1, l1, qs1)) (Rule (q2, l2, qs2)) =
  Rule ((q1, q2), l1, zip qs1 qs2);

datatype 'a dlist = Dlist of 'a list;

fun foldl f a [] = a
  | foldl f a (x :: xs) = foldl f (f a x) xs;

fun map_of A_ ((l, v) :: ps) k = (if eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

val ord_nat =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int ord;

fun the (SOME x) = x;

fun impl_of A_ (Rbt x) = x;

fun paint c Empty = Empty
  | paint c (Branch (uu, l, k, v, r)) = Branch (c, l, k, v, r);

fun balance (Branch (R, a, w, x, b)) s t (Branch (R, c, y, z, d)) =
  Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z Empty =
    Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, a, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z Empty =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance
    (Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c))) y z
    Empty =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance
    (Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance Empty w x (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, d))
  | balance Empty w x (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance Empty w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
  | balance Empty s t Empty = Branch (B, Empty, s, t, Empty)
  | balance Empty s t (Branch (B, va, vb, vc, vd)) =
    Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
  | balance Empty s t (Branch (v, Empty, vb, vc, Empty)) =
    Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
  | balance Empty s t (Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty)) =
    Branch
      (B, Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
  | balance Empty s t (Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi))) =
    Branch
      (B, Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance Empty s t
    (Branch (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
    = Branch
        (B, Empty, s, t,
          Branch
            (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance (Branch (B, va, vb, vc, vd)) s t Empty =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (B, ve, vf, vg, vh)) =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (v, Empty, vf, vg, Empty))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t,
          Branch
            (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (v, Empty, vb, vc, Empty)) s t Empty =
    Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
  | balance (Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh))) s t Empty =
    Branch
      (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty)
  | balance (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty)) s t Empty =
    Branch
      (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty)
  | balance
    (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)))
    s t Empty =
    Branch
      (B, Branch
            (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty)
  | balance (Branch (v, Empty, vf, vg, Empty)) s t (Branch (B, va, vb, vc, vd))
    = Branch
        (B, Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl))) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty)) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd))
  | balance
    (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)))
    s t (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch
            (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd));

fun rbt_ins A_ f k v Empty = Branch (R, Empty, k, v, Empty)
  | rbt_ins A_ f k v (Branch (B, l, x, y, r)) =
    (if less A_ k x then balance (rbt_ins A_ f k v l) x y r
      else (if less A_ x k then balance l x y (rbt_ins A_ f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_ins A_ f k v (Branch (R, l, x, y, r)) =
    (if less A_ k x then Branch (R, rbt_ins A_ f k v l, x, y, r)
      else (if less A_ x k then Branch (R, l, x, y, rbt_ins A_ f k v r)
             else Branch (R, l, x, f k y v, r)));

fun rbt_insert_with_key A_ f k v t = paint B (rbt_ins A_ f k v t);

fun rbt_insert A_ = rbt_insert_with_key A_ (fn _ => fn _ => fn nv => nv);

fun insert A_ k v t =
  Rbt (rbt_insert ((ord_preorder o preorder_order o order_linorder) A_) k v
        (impl_of A_ t));

fun rbt_lookup A_ Empty k = NONE
  | rbt_lookup A_ (Branch (uu, l, x, y, r)) k =
    (if less A_ k x then rbt_lookup A_ l k
      else (if less A_ x k then rbt_lookup A_ r k else SOME y));

fun lookup A_ t =
  rbt_lookup ((ord_preorder o preorder_order o order_linorder) A_)
    (impl_of A_ t);

fun cart it1 it2 empty3 ins3 s1 s2 =
  it1 s1 (fn _ => true) (fn x => it2 s2 (fn _ => true) (fn y => ins3 (x, y)))
    (empty3 ());

val emptya : 'a dlist = Dlist [];

fun member A_ [] y = false
  | member A_ (x :: xs) y = eq A_ x y orelse member A_ xs y;

fun inserta A_ x xs = (if member A_ xs x then xs else x :: xs);

fun list_of_dlist (Dlist x) = x;

fun insertb A_ x dxs = Dlist (inserta A_ x (list_of_dlist dxs));

fun membera A_ dxs = member A_ (list_of_dlist dxs);

fun remdups A_ [] = []
  | remdups A_ (x :: xs) =
    (if member A_ xs x then remdups A_ xs else x :: remdups A_ xs);

fun it_add update iti m1 m2 =
  iti m2 (fn _ => true) (fn (a, b) => update a b) m1;

val equal_nat = {equal = (fn a => fn b => ((a : IntInf.int) = b))} :
  IntInf.int equal;

val preorder_nat = {ord_preorder = ord_nat} : IntInf.int preorder;

val order_nat = {preorder_order = preorder_nat} : IntInf.int order;

fun list_all p [] = true
  | list_all p (x :: xs) = p x andalso list_all p xs;

fun is_none (SOME x) = false
  | is_none NONE = true;

fun iti_sel iti m f = iti m is_none (fn x => fn _ => f x) NONE;

fun it_size iti s = iti s (fn _ => true) (fn _ => suc) 0;

fun remap_rule f (Rule (q, l, qs)) = Rule (f q, l, map f qs);

type 'a hashable =
  {hashcode : 'a -> IntInf.int,
    bounded_hashcode : IntInf.int -> 'a -> IntInf.int,
    def_hashmap_size : 'a itself -> IntInf.int};
val hashcode = #hashcode : 'a hashable -> 'a -> IntInf.int;
val bounded_hashcode = #bounded_hashcode :
  'a hashable -> IntInf.int -> 'a -> IntInf.int;
val def_hashmap_size = #def_hashmap_size :
  'a hashable -> 'a itself -> IntInf.int;

fun list_all_zip p [] [] = true
  | list_all_zip p (a :: asa) (b :: bs) = p a b andalso list_all_zip p asa bs
  | list_all_zip p (v :: va) [] = false
  | list_all_zip p [] (v :: va) = false;

fun ls_memb A_ x l = membera A_ l x;

datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list;

datatype ('a, 'b) hashmap = RBT_HM of (IntInf.int, ('a, 'b) assoc_list) rbt;

datatype ('a, 'b, 'c) hashedTa_ext =
  HashedTa_ext of
    ('a, unit) hashmap * ('a, 'b) ta_rule dlist *
      ('b, ('a, 'b) ta_rule dlist) hashmap option *
      ('a, ('a, 'b) ta_rule dlist) hashmap option *
      (('a * 'b), ('a, 'b) ta_rule dlist) hashmap option * 'c;

fun hta_idx_f
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_idx_f;

val linorder_nat = {order_linorder = order_nat} : IntInf.int linorder;

fun rm_lookup A_ k m = lookup A_ m k;

fun impl_ofa (Assoc_List x) = x;

fun lookupa A_ al = map_of A_ (impl_ofa al);

fun lm_lookup A_ k m = lookupa A_ m k;

fun lookupb (A1_, A2_) k m =
  (case rm_lookup linorder_nat (hashcode A2_ k) m of NONE => NONE
    | SOME a => lm_lookup A1_ k a);

fun impl_of_RBT_HM A_ (RBT_HM x) = x;

fun hm_lookup (A1_, A2_) k hm = lookupb (A1_, A2_) k (impl_of_RBT_HM A2_ hm);

fun idx_lookup mlookup tempty i m =
  (case mlookup i m of NONE => tempty () | SOME s => s);

fun ls_empty x = (fn _ => emptya) x;

fun hl_idx_lookup (A1_, A2_) = idx_lookup (hm_lookup (A1_, A2_)) ls_empty;

fun hta_lookup_f (A1_, A2_) f h =
  hl_idx_lookup (A1_, A2_) f (the (hta_idx_f h));

fun ls_ins A_ = insertb A_;

fun it_image_filter iteratei1 empty2 ins2 f s =
  iteratei1 s (fn _ => true)
    (fn x => fn res => (case f x of NONE => res | SOME v => ins2 v res))
    (empty2 ());

fun foldli [] c f sigma = sigma
  | foldli (x :: xs) c f sigma =
    (if c sigma then foldli xs c f (f x sigma) else sigma);

fun dlist_iteratei xs = foldli (list_of_dlist xs);

fun ls_iteratei x = dlist_iteratei x;

fun ls_image_filter B_ = it_image_filter ls_iteratei ls_empty (ls_ins B_);

fun faccs A_ (B1_, B2_) h (Node (f, ts)) =
  let
    val qs = map (faccs A_ (B1_, B2_) h) ts;
  in
    ls_image_filter A_
      (fn r =>
        let
          val Rule (_, _, qsa) = r;
        in
          (if list_all_zip (fn q => fn qa => ls_memb A_ qa q) qs qsa
            then SOME (lhs r) else NONE)
        end)
      (hta_lookup_f (B1_, B2_) f h)
  end;

val emptyb : ('a, 'b) assoc_list = Assoc_List [];

fun lm_empty x = (fn _ => emptyb) x;

fun rm_update A_ = insert A_;

fun fst (a, b) = a;

fun snd (a, b) = b;

fun update_with_aux B_ v k f [] = [(k, f v)]
  | update_with_aux B_ v k f (p :: ps) =
    (if eq B_ (fst p) k then (k, f (snd p)) :: ps
      else p :: update_with_aux B_ v k f ps);

fun update_with A_ v k f al =
  Assoc_List (update_with_aux A_ v k f (impl_ofa al));

fun update A_ k v = update_with A_ v k (fn _ => v);

fun lm_update A_ = update A_;

fun updatea (A1_, A2_) k v m =
  let
    val hc = hashcode A2_ k;
  in
    (case rm_lookup linorder_nat hc m
      of NONE => rm_update linorder_nat hc (lm_update A1_ k v (lm_empty ())) m
      | SOME bm => rm_update linorder_nat hc (lm_update A1_ k v bm) m)
  end;

fun hm_update (A1_, A2_) k v hm =
  RBT_HM (updatea (A1_, A2_) k v (impl_of_RBT_HM A2_ hm));

fun rm_iterateoi Empty c f sigma = sigma
  | rm_iterateoi (Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_iterateoi l c f sigma;
           in
             (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa) else sigmaa)
           end
      else sigma);

fun rm_iterateoia A_ r = rm_iterateoi (impl_of A_ r);

fun rm_iteratei A_ = rm_iterateoia A_;

fun iteratei al c f = foldli (impl_ofa al) c f;

fun lm_iteratei x = iteratei x;

fun iterateia A_ m c f sigma_0 =
  rm_iteratei A_ m c (fn (_, lm) => lm_iteratei lm c f) sigma_0;

fun hm_iteratei A_ hm = iterateia linorder_nat (impl_of_RBT_HM A_ hm);

fun hm_add (A1_, A2_) = it_add (hm_update (A1_, A2_)) (hm_iteratei A2_);

fun rm_sel A_ = iti_sel (rm_iteratei A_);

fun lm_sel x = iti_sel lm_iteratei x;

fun sel A_ m f = rm_sel linorder_nat m (fn (_, lm) => lm_sel lm f);

fun hm_sel A_ hm = sel A_ (impl_of_RBT_HM A_ hm);

fun s_ins update x s = update x () s;

fun hs_ins (A1_, A2_) = s_ins (hm_update (A1_, A2_));

fun s_sel sel s f = sel s (fn (u, _) => f u);

fun hs_sel A_ = s_sel (hm_sel A_);

fun equal_lista A_ (a :: lista) [] = false
  | equal_lista A_ [] (a :: lista) = false
  | equal_lista A_ (aa :: listaa) (a :: lista) =
    eq A_ aa a andalso equal_lista A_ listaa lista
  | equal_lista A_ [] [] = true;

fun equal_list A_ = {equal = equal_lista A_} : ('a list) equal;

fun size_list [] = 0
  | size_list (a :: lista) = IntInf.+ (size_list lista, suc 0);

fun it_inter iteratei1 memb2 empt3 insrt3 s1 s2 =
  iteratei1 s1 (fn _ => true)
    (fn x => fn s => (if memb2 x s2 then insrt3 x s else s))
    (empt3 ());

fun it_union it ins s1 s2 = it s1 (fn _ => true) ins s2;

fun iti_ball iti s p = iti s (fn c => c) (fn x => fn _ => p x) true;

fun rm_empty A_ = (fn _ => empty A_);

fun emptyc A_ = rm_empty linorder_nat;

fun hm_empty_const A_ = RBT_HM (emptyc A_ ());

fun hm_empty A_ = (fn _ => hm_empty_const A_);

fun hs_empty A_ = hm_empty A_;

fun lh_image_filter (B1_, B2_) =
  it_image_filter ls_iteratei (hs_empty B2_) (hs_ins (B1_, B2_));

fun brc_iq (A1_, A2_) delta =
  lh_image_filter (A1_, A2_)
    (fn r => (if null (rhsq r) then SOME (lhs r) else NONE)) delta;

fun hta_Qi
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_Qi;

fun mod_nat m n = snd (IntInf.divMod (m, n));

fun fifo_empty x = (fn _ => ([], [])) x;

fun s_memb lookup x s = not (is_none (lookup x s));

fun hs_memb (A1_, A2_) = s_memb (hm_lookup (A1_, A2_));

fun s_iteratei it s c f = it s c (fn x => f (fst x));

fun hs_iteratei A_ = s_iteratei (hm_iteratei A_);

fun hs_size A_ = it_size (hs_iteratei A_);

fun ls_ball x = iti_ball ls_iteratei x;

fun ls_size x = it_size ls_iteratei x;

fun rs_iteratei A_ = s_iteratei (rm_iteratei A_);

fun rs_size A_ = it_size (rs_iteratei A_);

fun equal_ta_rulea A_ B_ (Rule (qa, la, listaa)) (Rule (q, l, lista)) =
  eq A_ qa q andalso (eq B_ la l andalso equal_lista A_ listaa lista);

fun equal_ta_rule A_ B_ = {equal = equal_ta_rulea A_ B_} :
  ('a, 'b) ta_rule equal;

fun ball_disjoint ball1 memb2 s1 s2 = ball1 s1 (fn x => not (memb2 x s2));

fun lh_disjoint (A1_, A2_) = ball_disjoint ls_ball (hs_memb (A1_, A2_));

fun hta_mema (A1_, A2_) (B1_, B2_) t h =
  not (lh_disjoint (B1_, B2_) (faccs B1_ (A1_, A2_) h t) (hta_Qi h));

fun hta_delta
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_delta;

fun hta_idx_f_update hta_idx_fa
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  HashedTa_ext
    (hta_Qi, hta_delta, hta_idx_fa hta_idx_f, hta_idx_s, hta_idx_sf, more);

fun idx_build_stepfun mlookup mupdate tempty tinsert f x m =
  let
    val i = f x;
  in
    (case mlookup i m of NONE => mupdate i (tinsert x (tempty ())) m
      | SOME s => mupdate i (tinsert x s) m)
  end;

fun idx_build mempty mlookup mupdate tempty tinsert siterate f s =
  siterate s (fn _ => true) (idx_build_stepfun mlookup mupdate tempty tinsert f)
    (mempty ());

fun hll_idx_build A_ (B1_, B2_) =
  idx_build (hm_empty B2_) (hm_lookup (B1_, B2_)) (hm_update (B1_, B2_))
    ls_empty (ls_ins A_) ls_iteratei;

fun build_rule_index A_ B_ (C1_, C2_) =
  hll_idx_build (equal_ta_rule A_ B_) (C1_, C2_);

fun build_rule_index_f A_ (B1_, B2_) delta =
  build_rule_index A_ B1_ (B1_, B2_) rhsl delta;

fun hta_ensure_idx_f A_ (B1_, B2_) h =
  (case hta_idx_f h
    of NONE =>
      hta_idx_f_update
        (fn _ => SOME (build_rule_index_f A_ (B1_, B2_) (hta_delta h))) h
    | SOME _ => h);

fun hta_mem (A1_, A2_) (B1_, B2_) t h =
  hta_mema (A1_, A2_) (B1_, B2_) t (hta_ensure_idx_f B1_ (A1_, A2_) h);

fun pa_cond A_ B_ C_ s = let
                           val (_, (w, _)) = s;
                         in
                           not (null w)
                         end;

fun hashcode_prod A_ B_ x =
  IntInf.+ (IntInf.* (hashcode A_ (fst x), (33 : IntInf.int)),
    hashcode B_ (snd x));

fun bounded_hashcode_prod A_ B_ n x =
  mod_nat
    (IntInf.+ (IntInf.* (bounded_hashcode A_ n (fst x), (33 : IntInf.int)),
      bounded_hashcode B_ n (snd x)))
    n;

fun def_hashmap_size_prod A_ B_ =
  (fn _ => IntInf.+ (def_hashmap_size A_ Type, def_hashmap_size B_ Type));

fun hashable_prod A_ B_ =
  {hashcode = hashcode_prod A_ B_,
    bounded_hashcode = bounded_hashcode_prod A_ B_,
    def_hashmap_size = def_hashmap_size_prod A_ B_}
  : ('a * 'b) hashable;

fun equal_proda A_ B_ (aa, ba) (a, b) = eq A_ aa a andalso eq B_ ba b;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) equal;

fun pa_upd_rule (A1_, A2_) (B1_, B2_) q w [] = (q, w)
  | pa_upd_rule (A1_, A2_) (B1_, B2_) q w (qp :: qs) =
    (if not (hs_memb (equal_prod A1_ B1_, hashable_prod A2_ B2_) qp q)
      then pa_upd_rule (A1_, A2_) (B1_, B2_)
             (hs_ins (equal_prod A1_ B1_, hashable_prod A2_ B2_) qp q) (qp :: w)
             qs
      else pa_upd_rule (A1_, A2_) (B1_, B2_) q w qs);

fun hta_idx_s
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_idx_s;

fun hta_lookup_s (A1_, A2_) q h =
  hl_idx_lookup (A1_, A2_) q (the (hta_idx_s h));

fun ls_ins_dj A_ = insertb A_;

fun hta_idx_sf
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_idx_sf;

fun hta_lookup_sf (A1_, A2_) (B1_, B2_) q f h =
  hl_idx_lookup (equal_prod A1_ B1_, hashable_prod A2_ B2_) (q, f)
    (the (hta_idx_sf h));

fun pa_step (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2 s =
  let
    val (q, (w, delta_d)) = s;
    val (q1, q2) = hd w;
  in
    ls_iteratei (hta_lookup_s (A1_, A2_) q1 h1) (fn _ => true)
      (fn r1 =>
        ls_iteratei (hta_lookup_sf (C1_, C2_) (B1_, B2_) q2 (rhsl r1) h2)
          (fn _ => true)
          (fn r2 => fn res =>
            (if (((size_list (rhsq r1)) : IntInf.int) = (size_list (rhsq r2)))
              then let
                     val rp = r_prod r1 r2;
                     val (qa, (wa, delta_da)) = res;
                     val (qb, wb) =
                       pa_upd_rule (A1_, A2_) (C1_, C2_) qa wa (rhsq rp);
                   in
                     (qb, (wb, ls_ins_dj
                                 (equal_ta_rule (equal_prod A1_ C1_) B1_) rp
                                 delta_da))
                   end
              else res)))
      (q, (tl w, delta_d))
  end;

fun map_to_nat iterate1 empty2 update2 s =
  snd (iterate1 s (fn _ => true)
         (fn x => fn (c, m) => (IntInf.+ (c, (1 : IntInf.int)), update2 x c m))
        (0, empty2 ()));

fun equal_dlista A_ dxs dys =
  equal_lista A_ (list_of_dlist dxs) (list_of_dlist dys);

fun equal_dlist A_ = {equal = equal_dlista A_} : 'a dlist equal;

fun iflt_image iflt f s = iflt (fn x => SOME (f x)) s;

fun hs_image_filter A_ (B1_, B2_) =
  it_image_filter (hs_iteratei A_) (hs_empty B2_) (hs_ins (B1_, B2_));

fun hs_image A_ (B1_, B2_) = iflt_image (hs_image_filter A_ (B1_, B2_));

fun s_ins_dj update_dj x s = update_dj x () s;

fun hm_update_dj (A1_, A2_) = hm_update (A1_, A2_);

fun hs_ins_dj (A1_, A2_) = s_ins_dj (hm_update_dj (A1_, A2_));

fun hs_inter (A1_, A2_) =
  it_inter (hs_iteratei A2_) (hs_memb (A1_, A2_)) (hs_empty A2_)
    (hs_ins_dj (A1_, A2_));

fun hs_union (A1_, A2_) = hm_add (A1_, A2_);

fun hhh_cart (A1_, A2_) (B1_, B2_) =
  cart (hs_iteratei A2_) (hs_iteratei B2_) (hs_empty (hashable_prod A2_ B2_))
    (hs_ins_dj (equal_prod A1_ B1_, hashable_prod A2_ B2_));

fun brc_cond A_ B_ = (fn (_, (w, _)) => not (null w));

fun rqrm_lookup (A1_, A2_) rqrm q =
  (case hm_lookup (A1_, A2_) q rqrm of NONE => ls_empty () | SOME s => s);

fun hashcode_list A_ =
  foldl (fn h => fn x =>
          IntInf.+ (IntInf.* (h, (33 : IntInf.int)), hashcode A_ x))
    (5381 : IntInf.int);

fun hashcode_of_ta_rule A_ B_ (Rule (q, f, qs)) =
  IntInf.+ (IntInf.+ (hashcode A_ q, hashcode B_ f), hashcode_list A_ qs);

fun hashcode_ta_rule A_ B_ = hashcode_of_ta_rule A_ B_;

fun bounded_hashcode_ta_rule A_ B_ n r =
  mod_nat (hashcode_of_ta_rule A_ B_ r) n;

fun def_hashmap_size_ta_rule A_ B_ = (fn _ => (32 : IntInf.int));

fun hashable_ta_rule A_ B_ =
  {hashcode = hashcode_ta_rule A_ B_,
    bounded_hashcode = bounded_hashcode_ta_rule A_ B_,
    def_hashmap_size = def_hashmap_size_ta_rule A_ B_}
  : ('a, 'b) ta_rule hashable;

fun brc_inner_step (A1_, A2_) (B1_, B2_) r =
  (fn (q, (w, rcm)) =>
    let
      val c =
        the (hm_lookup (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r rcm);
      val rcma =
        hm_update (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r
          (IntInf.max (0, IntInf.- (c, (1 : IntInf.int)))) rcm;
      val qa =
        (if IntInf.<= (c, (1 : IntInf.int)) then hs_ins (A1_, A2_) (lhs r) q
          else q);
      val wa =
        (if IntInf.<= (c, (1 : IntInf.int)) andalso
              not (hs_memb (A1_, A2_) (lhs r) q)
          then lhs r :: w else w);
    in
      (qa, (wa, rcma))
    end);

fun brc_step (A1_, A2_) (B1_, B2_) rqrm =
  (fn (q, (w, rcm)) =>
    ls_iteratei (rqrm_lookup (A1_, A2_) rqrm (hd w)) (fn _ => true)
      (brc_inner_step (A1_, A2_) (B1_, B2_)) (q, (tl w, rcm)));

fun brec_iqm (A1_, A2_) B_ delta =
  ls_iteratei delta (fn _ => true)
    (fn r => fn m =>
      (if null (rhsq r) then hm_update (A1_, A2_) (lhs r) (Node (rhsl r, [])) m
        else m))
    (hm_empty A2_ ());

fun init_hta qi delta = HashedTa_ext (qi, delta, NONE, NONE, NONE, ());

fun it_set_to_list iti s = iti s (fn _ => true) (fn a => fn b => a :: b) [];

fun hs_to_list A_ = it_set_to_list (hs_iteratei A_);

fun pa_initial (A1_, A2_) B_ (C1_, C2_) h1 h2 =
  let
    val qip = hhh_cart (A1_, A2_) (C1_, C2_) (hta_Qi h1) (hta_Qi h2);
  in
    (qip, (hs_to_list (hashable_prod A2_ C2_) qip, ls_empty ()))
  end;

fun while_option b c s = (if b s then while_option b c (c s) else SOME s);

fun whilea b c s = the (while_option b c s);

fun hta_proda (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2 =
  let
    val (_, (_, a)) =
      whilea (pa_cond A2_ C2_ B2_)
        (pa_step (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2)
        (pa_initial (A1_, A2_) B2_ (C1_, C2_) h1 h2);
  in
    init_hta (hhh_cart (A1_, A2_) (C1_, C2_) (hta_Qi h1) (hta_Qi h2)) a
  end;

fun hta_idx_s_update hta_idx_sa
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  HashedTa_ext
    (hta_Qi, hta_delta, hta_idx_f, hta_idx_sa hta_idx_s, hta_idx_sf, more);

fun build_rule_index_s (A1_, A2_) B_ delta =
  build_rule_index A1_ B_ (A1_, A2_) lhs delta;

fun hta_ensure_idx_s (A1_, A2_) B_ h =
  (case hta_idx_s h
    of NONE =>
      hta_idx_s_update
        (fn _ => SOME (build_rule_index_s (A1_, A2_) B_ (hta_delta h))) h
    | SOME _ => h);

fun hta_idx_sf_update hta_idx_sfa
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  HashedTa_ext
    (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sfa hta_idx_sf, more);

fun build_rule_index_sf (A1_, A2_) (B1_, B2_) delta =
  build_rule_index A1_ B1_ (equal_prod A1_ B1_, hashable_prod A2_ B2_)
    (fn r => (lhs r, rhsl r)) delta;

fun hta_ensure_idx_sf (A1_, A2_) (B1_, B2_) h =
  (case hta_idx_sf h
    of NONE =>
      hta_idx_sf_update
        (fn _ => SOME (build_rule_index_sf (A1_, A2_) (B1_, B2_) (hta_delta h)))
        h
    | SOME _ => h);

fun hta_prod (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2 =
  hta_proda (A1_, A2_) (B1_, B2_) (C1_, C2_)
    (hta_ensure_idx_s (A1_, A2_) B1_ h1)
    (hta_ensure_idx_sf (C1_, C2_) (B1_, B2_) h2);

fun hashcode_nat n = n;

fun bounded_hashcode_nat na n = mod_nat n na;

fun def_hashmap_size_nat x = (fn _ => (16 : IntInf.int)) x;

val hashable_nat =
  {hashcode = hashcode_nat, bounded_hashcode = bounded_hashcode_nat,
    def_hashmap_size = def_hashmap_size_nat}
  : IntInf.int hashable;

fun htai_mem x = hta_mem (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

fun rqrm_add (A1_, A2_) B_ q r res =
  (case hm_lookup (A1_, A2_) q res
    of NONE => hm_update (A1_, A2_) q (ls_ins B_ r (ls_empty ())) res
    | SOME s => hm_update (A1_, A2_) q (ls_ins B_ r s) res);

fun fifo_dequeue f =
  (case snd f of [] => let
                         val rp = rev (fst f);
                       in
                         (hd rp, ([], tl rp))
                       end
    | a :: l => (a, (fst f, l)));

fun fifo_enqueue a f = (a :: fst f, snd f);

fun fifo_isEmpty A_ f = equal_proda (equal_list A_) (equal_list A_) f ([], []);

fun lhh_union (A1_, A2_) = it_union ls_iteratei (hs_ins (A1_, A2_));

fun brec_cond A_ =
  (fn (_, (w, (_, qwit))) => not (fifo_isEmpty A_ w) andalso is_none qwit);

fun brec_construct_witness (A1_, A2_) B_ qm r =
  Node (rhsl r, map (fn q => the (hm_lookup (A1_, A2_) q qm)) (rhsq r));

fun brec_inner_step (A1_, A2_) (B1_, B2_) qi r =
  (fn (q, (w, (rcm, qwit))) =>
    let
      val c =
        the (hm_lookup (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r rcm);
      val cond =
        IntInf.<= (c, (1 : IntInf.int)) andalso
          is_none (hm_lookup (A1_, A2_) (lhs r) q);
      val rcma =
        hm_update (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r
          (IntInf.max (0, IntInf.- (c, (1 : IntInf.int)))) rcm;
      val qa =
        (if cond
          then hm_update (A1_, A2_) (lhs r)
                 (brec_construct_witness (A1_, A2_) B2_ q r) q
          else q);
      val wa = (if cond then fifo_enqueue (lhs r) w else w);
      val qwita =
        (if IntInf.<= (c, (1 : IntInf.int)) andalso
              hs_memb (A1_, A2_) (lhs r) qi
          then SOME (lhs r) else qwit);
    in
      (qa, (wa, (rcma, qwita)))
    end);

fun brec_step (A1_, A2_) (B1_, B2_) rqrm qi =
  (fn (q, (w, (rcm, qwit))) =>
    let
      val (qa, wa) = fifo_dequeue w;
    in
      ls_iteratei (rqrm_lookup (A1_, A2_) rqrm qa) (fn _ => true)
        (brec_inner_step (A1_, A2_) (B1_, B2_) qi) (q, (wa, (rcm, qwit)))
    end);

fun hta_empty A_ B_ u = init_hta (hs_empty A_ ()) (ls_empty ());

fun ls_image B_ = iflt_image (ls_image_filter B_);

fun hta_remap A_ (B1_, B2_) (C1_, C2_) f h =
  init_hta (hs_image A_ (B1_, B2_) f (hta_Qi h))
    (ls_image (equal_ta_rule B1_ C1_) (remap_rule f) (hta_delta h));

fun equal_ustate_wrappera A_ B_ (USW2 q2) (USW1 q1) = false
  | equal_ustate_wrappera A_ B_ (USW1 q1) (USW2 q2) = false
  | equal_ustate_wrappera A_ B_ (USW2 q2a) (USW2 q2) = eq B_ q2a q2
  | equal_ustate_wrappera A_ B_ (USW1 q1a) (USW1 q1) = eq A_ q1a q1;

fun equal_ustate_wrapper A_ B_ = {equal = equal_ustate_wrappera A_ B_} :
  ('a, 'b) ustate_wrapper equal;

fun ls_union A_ = it_union ls_iteratei (ls_ins A_);

fun ls_union_dj A_ = ls_union A_;

fun hashcode_ustate_wrapper A_ B_ x =
  (case x of USW1 a => IntInf.* ((2 : IntInf.int), hashcode A_ a)
    | USW2 b =>
      IntInf.+ (IntInf.* ((2 : IntInf.int), hashcode B_ b), (1 : IntInf.int)));

fun bounded_hashcode_ustate_wrapper A_ B_ n x =
  (case x of USW1 a => bounded_hashcode A_ n a
    | USW2 b =>
      IntInf.max (0, IntInf.- (IntInf.max (0, IntInf.- (n, (1 : IntInf.int))),
        bounded_hashcode B_ n b)));

fun def_hashmap_size_ustate_wrapper A_ B_ =
  (fn _ => IntInf.+ (def_hashmap_size A_ Type, def_hashmap_size B_ Type));

fun hashable_ustate_wrapper A_ B_ =
  {hashcode = hashcode_ustate_wrapper A_ B_,
    bounded_hashcode = bounded_hashcode_ustate_wrapper A_ B_,
    def_hashmap_size = def_hashmap_size_ustate_wrapper A_ B_}
  : ('a, 'b) ustate_wrapper hashable;

fun hta_union (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2 =
  init_hta
    (hs_union (equal_ustate_wrapper A1_ C1_, hashable_ustate_wrapper A2_ C2_)
      (hs_image A2_
        (equal_ustate_wrapper A1_ C1_, hashable_ustate_wrapper A2_ C2_) USW1
        (hta_Qi h1))
      (hs_image C2_
        (equal_ustate_wrapper A1_ C1_, hashable_ustate_wrapper A2_ C2_) USW2
        (hta_Qi h2)))
    (ls_union_dj (equal_ta_rule (equal_ustate_wrapper A1_ C1_) B1_)
      (ls_image (equal_ta_rule (equal_ustate_wrapper A1_ C1_) B1_)
        (remap_rule USW1) (hta_delta h1))
      (ls_image (equal_ta_rule (equal_ustate_wrapper A1_ C1_) B1_)
        (remap_rule USW2) (hta_delta h2)));

fun list_to_ls A_ xs = Dlist (remdups A_ xs);

fun rule_states_l A_ (Rule (q, f, qs)) = ls_ins A_ q (list_to_ls A_ qs);

fun it_Union_image iti1 em3 un233 f s =
  iti1 s (fn _ => true) (fn x => un233 (f x)) (em3 ());

fun llh_Union_image (B1_, B2_) =
  it_Union_image ls_iteratei (hs_empty B2_) (lhh_union (B1_, B2_));

fun hta_delta_states (A1_, A2_) h =
  llh_Union_image (A1_, A2_) id
    (ls_image_filter (equal_dlist A1_) (fn r => SOME (rule_states_l A1_ r))
      (hta_delta h));

fun hta_states (A1_, A2_) h =
  hs_union (A1_, A2_) (hta_Qi h) (hta_delta_states (A1_, A2_) h);

fun hh_map_to_nat (A1_, A2_) =
  map_to_nat (hs_iteratei A2_) (hm_empty A2_) (hm_update (A1_, A2_));

fun hta_reindex (A1_, A2_) (B1_, B2_) h =
  let
    val mp = hh_map_to_nat (A1_, A2_) (hta_states (A1_, A2_) h);
  in
    hta_remap A2_ (equal_nat, hashable_nat) (B1_, B2_)
      (fn q => the (hm_lookup (A1_, A2_) q mp)) h
  end;

fun htai_prod h1 h2 =
  hta_reindex
    (equal_prod equal_nat equal_nat, hashable_prod hashable_nat hashable_nat)
    (equal_nat, hashable_nat)
    (hta_prod (equal_nat, hashable_nat) (equal_nat, hashable_nat)
      (equal_nat, hashable_nat) h1 h2);

fun it_set_to_List_enq iterate emp enq s =
  iterate s (fn _ => true) enq (emp ());

fun it_set_to_fifo it = it_set_to_List_enq it fifo_empty fifo_enqueue;

fun hs_to_fifo A_ = it_set_to_fifo (hs_iteratei A_);

fun build_rqrm (A1_, A2_) (B1_, B2_) delta =
  ls_iteratei delta (fn _ => true)
    (fn r => fn res =>
      foldl (fn resa => fn q =>
              rqrm_add (A1_, A2_) (equal_ta_rule A1_ B1_) q r resa)
        res (rhsq r))
    (hm_empty A2_ ());

fun hta_add_qi (A1_, A2_) B_ qi h =
  init_hta (hs_ins (A1_, A2_) qi (hta_Qi h)) (hta_delta h);

fun image_filter_cartesian_product iterate1 iterate2 empty3 insert3 f s1 s2 =
  iterate1 s1 (fn _ => true)
    (fn x =>
      iterate2 s2 (fn _ => true)
        (fn y => fn res =>
          (case f (x, y) of NONE => res | SOME z => insert3 z res)))
    (empty3 ());

fun image_filter_cp iterate1 iterate2 empty3 insert3 f p s1 s2 =
  image_filter_cartesian_product iterate1 iterate2 empty3 insert3
    (fn xy => (if p xy then SOME (f xy) else NONE)) s1 s2;

fun lll_inj_image_filter_cp C_ =
  image_filter_cp ls_iteratei ls_iteratei ls_empty (ls_ins_dj C_);

fun delta_prod_h (A1_, A2_) (B1_, B2_) (C1_, C2_) delta_1 delta_2 =
  lll_inj_image_filter_cp (equal_ta_rule (equal_prod A1_ C1_) B1_)
    (fn (a, b) => r_prod a b)
    (fn (r1, r2) =>
      eq B1_ (rhsl r1) (rhsl r2) andalso
        (((size_list (rhsq r1)) : IntInf.int) = (size_list (rhsq r2))))
    delta_1 delta_2;

fun hta_prodWR (A1_, A2_) (B1_, B2_) (D1_, D2_) h1 h2 =
  init_hta (hhh_cart (A1_, A2_) (D1_, D2_) (hta_Qi h1) (hta_Qi h2))
    (delta_prod_h (A1_, A2_) (B1_, B2_) (D1_, D2_) (hta_delta h1)
      (hta_delta h2));

fun hta_reduce (A1_, A2_) B_ h q =
  init_hta (hs_inter (A1_, A2_) q (hta_Qi h))
    (ls_image_filter (equal_ta_rule A1_ B_)
      (fn r =>
        (if hs_memb (A1_, A2_) (lhs r) q andalso
              list_all (fn qa => hs_memb (A1_, A2_) qa q) (rhsq r)
          then SOME r else NONE))
      (hta_delta h));

fun htai_empty x = hta_empty hashable_nat hashable_nat x;

fun htai_union h1 h2 =
  hta_reindex
    (equal_ustate_wrapper equal_nat equal_nat,
      hashable_ustate_wrapper hashable_nat hashable_nat)
    (equal_nat, hashable_nat)
    (hta_union (equal_nat, hashable_nat) (equal_nat, hashable_nat)
      (equal_nat, hashable_nat) h1 h2);

fun brc_rcm_init (A1_, A2_) (B1_, B2_) delta =
  ls_iteratei delta (fn _ => true)
    (fn r =>
      hm_update (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r
        (size_list (remdups A1_ (rhsq r))))
    (hm_empty (hashable_ta_rule A2_ B2_) ());

fun brc_initial (A1_, A2_) (B1_, B2_) delta =
  let
    val iq = brc_iq (A1_, A2_) delta;
  in
    (iq, (hs_to_list A2_ iq, brc_rcm_init (A1_, A2_) (B1_, B2_) delta))
  end;

fun htai_add_qi x = hta_add_qi (equal_nat, hashable_nat) hashable_nat x;

fun htai_prodWR h1 h2 =
  hta_reindex
    (equal_prod equal_nat equal_nat, hashable_prod hashable_nat hashable_nat)
    (equal_nat, hashable_nat)
    (hta_prodWR (equal_nat, hashable_nat) (equal_nat, hashable_nat)
      (equal_nat, hashable_nat) h1 h2);

fun sel_disjoint_witness sel1 mem2 s1 s2 =
  sel1 s1 (fn x => (if mem2 x s2 then SOME x else NONE));

fun hh_disjoint_witness (A1_, A2_) =
  sel_disjoint_witness (hs_sel A2_) (hs_memb (A1_, A2_));

fun brec_initial (A1_, A2_) (B1_, B2_) qi delta =
  let
    val iq = brc_iq (A1_, A2_) delta;
  in
    (brec_iqm (A1_, A2_) B2_ delta,
      (hs_to_fifo A2_ iq,
        (brc_rcm_init (A1_, A2_) (B1_, B2_) delta,
          hh_disjoint_witness (A1_, A2_) iq qi)))
  end;

fun hta_add_rule (A1_, A2_) (B1_, B2_) r h =
  init_hta (hta_Qi h) (ls_ins (equal_ta_rule A1_ B1_) r (hta_delta h));

fun htai_add_rule x =
  hta_add_rule (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

fun hta_bwd_reduce (A1_, A2_) (B1_, B2_) h =
  let
    val rqrm = build_rqrm (A1_, A2_) (B1_, B2_) (hta_delta h);
  in
    hta_reduce (A1_, A2_) B1_ h
      (fst (whilea (brc_cond A2_ B2_) (brc_step (A1_, A2_) (B1_, B2_) rqrm)
             (brc_initial (A1_, A2_) (B1_, B2_) (hta_delta h))))
  end;

fun htai_bwd_reduce x =
  hta_bwd_reduce (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

fun htai_ensure_idx_f x =
  hta_ensure_idx_f equal_nat (equal_nat, hashable_nat) x;

fun htai_ensure_idx_s x =
  hta_ensure_idx_s (equal_nat, hashable_nat) equal_nat x;

fun htai_ensure_idx_sf x =
  hta_ensure_idx_sf (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

fun hta_is_empty_witness (A1_, A2_) (B1_, B2_) h =
  let
    val rqrm = build_rqrm (A1_, A2_) (B1_, B2_) (hta_delta h);
  in
    (case whilea (brec_cond A1_)
            (brec_step (A1_, A2_) (B1_, B2_) rqrm (hta_Qi h))
            (brec_initial (A1_, A2_) (B1_, B2_) (hta_Qi h) (hta_delta h))
      of (q, (_, (_, NONE))) => NONE
      | (q, (_, (_, SOME qa))) => hm_lookup (A1_, A2_) qa q)
  end;

fun htai_is_empty_witness x =
  hta_is_empty_witness (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

end; (*struct Ta*)
